from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TestModel',
            fields=[
                ('name', models.TextField(primary_key=True, serialize=False, verbose_name='Name')),
                ('isAdmin', models.BooleanField(verbose_name='Is Admin')),
                ('principal_identifier', models.TextField(verbose_name='Principal identifier')),
            ],
        ),
    ]



from django.test import TestCase
from identitylib.identifiers import Identifier, IdentifierSchemes
from rest_framework.test import APIRequestFactory
from rest_framework.exceptions import AuthenticationFailed

from apigatewayauth.api_gateway_auth import (
    APIGatewayAuthentication, APIGatewayAuthenticationDetails
)


class APIGatewayAuthTestCase(TestCase):

    def setUp(self):
        super().setUp()

        self.request_factory = APIRequestFactory()
        self.auth = APIGatewayAuthentication()

    def request_with_headers(self, headers={}):
        parsed_headers = {
            f'HTTP_{key.upper().replace("-", "_")}': value
            for key, value in headers.items()
        }
        return self.request_factory.get('/', **parsed_headers)

    def test_bails_early_without_api_org(self):
        self.assertIsNone(
            self.auth.authenticate(self.request_with_headers({"Accept": "application/json"}))
        )

    def test_throws_without_auth_details(self):
        with self.assertRaisesMessage(
            AuthenticationFailed, 'Could not authenticate using x-api-* headers'
        ):
            self.auth.authenticate(self.request_with_headers({"x-api-org-name": "test"}))

    def test_throws_without_principal_identifier(self):
        with self.assertRaisesMessage(
            AuthenticationFailed, 'Could not authenticate using x-api-* headers'
        ):
            self.auth.authenticate(self.request_with_headers({
                "x-api-org-name": "test",
                "x-api-developer-app-class": "public"
            }))

    def test_throws_with_bad_principal_identifier(self):
        with self.assertRaisesMessage(
            AuthenticationFailed, 'Invalid principal identifier'
        ):
            self.auth.authenticate(self.request_with_headers({
                "x-api-org-name": "test",
                "x-api-developer-app-class": "public",
                "x-api-oauth2-user": "Monty Dawson"
            }))

    def test_can_use_any_identifier_scheme_in_principal_identifier(self):
        for scheme in IdentifierSchemes.get_registered_schemes():
            _, auth = self.auth.authenticate(self.request_with_headers({
                "x-api-org-name": "test",
                "x-api-developer-app-class": "public",
                "x-api-oauth2-user": str(Identifier("1000", scheme))
            }))
            self.assertEqual(auth.principal_identifier, Identifier("1000", scheme))

    def test_throws_with_unknown_identifier_type(self):
        with self.assertRaisesMessage(
            AuthenticationFailed, 'Invalid principal identifier'
        ):
            self.auth.authenticate(self.request_with_headers({
                "x-api-org-name": "test",
                "x-api-developer-app-class": "public",
                "x-api-oauth2-user": 'wgd23@gmail.com'
            }))

    def test_returns_client_details_for_valid_auth(self):
        user, auth = self.auth.authenticate(self.request_with_headers({
            "x-api-org-name": "test",
            "x-api-developer-app-class": "public",
            "x-api-oauth2-user": str(Identifier('a123', IdentifierSchemes.CRSID))
        }))
        self.assertIsNone(user)

        self.assertEqual(
            auth,
            APIGatewayAuthenticationDetails(
                Identifier('a123', IdentifierSchemes.CRSID),
                set(),
                None,
                None,
            )
        )

    def test_will_pass_through_scopes(self):
        _, auth = self.auth.authenticate(self.request_with_headers({
            "x-api-org-name": "test",
            "x-api-developer-app-class": "public",
            "x-api-oauth2-user": str(Identifier('a123', IdentifierSchemes.CRSID)),
            "x-api-oauth2-scope": (
                "https://api.apps.cam.ac.uk/a.readonly https://api.apps.cam.ac.uk/b"
            )
        }))

        self.assertEqual(
            auth,
            APIGatewayAuthenticationDetails(
                Identifier('a123', IdentifierSchemes.CRSID),
                set(['https://api.apps.cam.ac.uk/a.readonly', 'https://api.apps.cam.ac.uk/b']),
                None,
                None,
            ),
        )

    def test_will_pass_through_app_and_client_ids(self):
        _, auth = self.auth.authenticate(self.request_with_headers({
            "x-api-org-name": "test",
            "x-api-developer-app-class": "confidential",
            "x-api-oauth2-user": str(Identifier('a123', IdentifierSchemes.CRSID)),
            "x-api-oauth2-scope": (
                "https://api.apps.cam.ac.uk/a.readonly https://api.apps.cam.ac.uk/b"
            ),
            "x-api-developer-app-id": "app-uuid-mock",
            "x-api-oauth2-client-id": "client-id-uuid-mock",
        }))

        self.assertEqual(
            auth,
            APIGatewayAuthenticationDetails(
                Identifier('a123', IdentifierSchemes.CRSID),
                set(['https://api.apps.cam.ac.uk/a.readonly', 'https://api.apps.cam.ac.uk/b']),
                'app-uuid-mock',
                'client-id-uuid-mock'
            )
        )



from unittest.mock import patch
from django.test import TestCase
from django.conf import settings
from django.core.cache import cache
from identitylib.identifiers import Identifier, IdentifierSchemes

from apigatewayauth.permissions_spec import (
    get_permission_spec, get_principals_with_permission, get_groups_with_permission
)


class PermissionSpecTestCase(TestCase):

    def setUp(self):
        super().setUp()
        cache.clear()  # clear the cache between tests

    @patch('apigatewayauth.permissions_spec.geddit')
    def test_will_return_parsed_permissions_spec_with_cache(self, geddit_mock):
        geddit_mock.return_value = """
            CARD_DATA_READERS:
                principals:
                    - abc123@v1.person.identifiers.cam.ac.uk
                    - 1234@application.api.apps.cam.ac.uk
                groups:
                    - 1001@groups.lookup.cam.ac.uk
        """

        expected_permission_spec = {
            "CARD_DATA_READERS": {
                "principals": [
                    str(Identifier('abc123', IdentifierSchemes.CRSID)),
                    str(Identifier('1234', IdentifierSchemes.API_GATEWAY_APPLICATION)),
                ],
                "groups": [
                    str(Identifier('1001', IdentifierSchemes.LOOKUP_GROUP)),
                ],
            }
        }
        self.assertEqual(get_permission_spec(), expected_permission_spec)
        geddit_mock.assert_called_with(settings.PERMISSIONS_SPECIFICATION_URL)

        geddit_mock.reset_mock()

        # ensure that getting the permissions spec again does not use geddit as the cached
        # result is used
        self.assertEqual(get_permission_spec(), expected_permission_spec)
        geddit_mock.assert_not_called()

    @patch('apigatewayauth.permissions_spec.geddit')
    def test_can_query_specific_permission(self, geddit_mock):
        geddit_mock.return_value = """
            CARD_DATA_READERS:
                principals:
                    - 1234@application.api.apps.cam.ac.uk
                    - abc123@v1.person.identifiers.cam.ac.uk
                groups:
                    - 1001@groups.lookup.cam.ac.uk
            CARD_DATA_WRITERS:
                principals:
                    - abc234@v1.person.identifiers.cam.ac.uk
        """

        self.assertEqual(
            get_principals_with_permission('CARD_DATA_READERS'), {
                Identifier('1234', IdentifierSchemes.API_GATEWAY_APPLICATION),
                Identifier('abc123', IdentifierSchemes.CRSID)
            }
        )
        geddit_mock.assert_called_with(settings.PERMISSIONS_SPECIFICATION_URL)
        geddit_mock.reset_mock()

        self.assertEqual(
            get_groups_with_permission('CARD_DATA_READERS'), {
                Identifier('1001', IdentifierSchemes.LOOKUP_GROUP)
            }
        )
        # should not be called as we have cached the spec
        geddit_mock.assert_not_called()

        self.assertEqual(
            get_principals_with_permission('CARD_DATA_WRITERS'), {
                Identifier('abc234', IdentifierSchemes.CRSID)
            }
        )

        self.assertEqual(get_groups_with_permission('CARD_DATA_WRITERS'), set())

        self.assertEqual(get_groups_with_permission('CARD_DATA_ADMINS'), set())
        self.assertEqual(get_principals_with_permission('CARD_DATA_ADMINS'), set())



from typing import Optional, Set
from dataclasses import dataclass
from rest_framework.request import Request
from identitylib.identifiers import Identifier

from rest_framework import authentication
from rest_framework.exceptions import AuthenticationFailed


@dataclass(eq=True)
class APIGatewayAuthenticationDetails:
    """
    A dataclass representing the authentication information passed from the API Gateway.

    """

    principal_identifier: Identifier
    scopes: Set[str]
    app_id: Optional[str] = None
    client_id: Optional[str] = None


class APIGatewayAuthentication(authentication.BaseAuthentication):
    """
    An Authentication provider which interprets the headers provided by the API Gateway.

    This library expects to only be used within an application that is deployed behind and can
    only be invoked by the API Gateway, and therefore relies on the fact that the headers
    provided are authoritative.

    """

    def authenticate(self, request: Request):
        if not request.META.get('HTTP_X_API_ORG_NAME', None):
            # bail early if we look like we're not being called by the API Gateway
            return None
        if not request.META.get('HTTP_X_API_OAUTH2_USER', None):
            raise AuthenticationFailed('Could not authenticate using x-api-* headers')

        try:
            principal_identifier = Identifier.from_string(
                request.META['HTTP_X_API_OAUTH2_USER'],
                find_by_alias=True
            )
        except Exception:
            raise AuthenticationFailed('Invalid principal identifier')

        auth = APIGatewayAuthenticationDetails(
            principal_identifier=principal_identifier,
            scopes=set(filter(bool, request.META.get('HTTP_X_API_OAUTH2_SCOPE', '').split(' '))),
            # the following will only be populated for confidential clients
            app_id=request.META.get('HTTP_X_API_DEVELOPER_APP_ID', None),
            client_id=request.META.get('HTTP_X_API_OAUTH2_CLIENT_ID', None)
        )
        # the first item in the tuple represents the 'user' which we don't have when we've
        # used the API Gateway for authentication.
        return None, auth



from typing import Dict
from functools import wraps
from tempfile import NamedTemporaryFile
import yaml

from django.test import override_settings


def override_permission_spec(permissions_spec: Dict[str, Dict[str, str]]):
    """
    A decorator which allows the permissions specification to be mocked, allowing a
    permission to only be enabled for the given identities.

    """
    def decorator(func):
        @wraps(func)
        def wrapped_function(*args, **kwargs):
            with NamedTemporaryFile('w+') as temp_file:
                yaml.dump(permissions_spec, temp_file.file)
                with override_settings(PERMISSIONS_SPECIFICATION_URL=temp_file.name):
                    func(*args, **kwargs)
        return wrapped_function
    return decorator



from typing import Set
from django.core.cache import cache
from rest_framework import permissions, request
from identitylib.identifiers import IdentifierSchemes

from ucamlookup.utils import get_connection
from ucamlookup.ibisclient import PersonMethods, IbisException

from logging import getLogger

from .api_gateway_auth import APIGatewayAuthenticationDetails
from .permissions_spec import (
    get_groups_with_permission, get_permission_spec, get_principals_with_permission
)


LOG = getLogger(__name__)


class Disallowed(permissions.BasePermission):
    """
    A permissions class which disallows all access, this is used as the default permissions
    class to stop routes being added which accidentally expose data.

    """
    def has_permission(self, request, view):
        return False

    def has_object_permission(self, request, view, obj):
        return False


class IsResourceOwningPrincipal(permissions.BasePermission):
    """
    A permissions class which ensures that the client represents a principal who is limited
    to accessing only resources they own.

    """

    message = 'Please authenticate as the owning user using the API Gateway.'

    @staticmethod
    def get_queryset_for_principal(request, base_object):
        """
        A helper method which filters the given object's queryset to the owning principal
        if required

        """
        if not getattr(request, 'should_limit_to_resource_owning_principal', False):
            return base_object.objects.all()
        if not isinstance(getattr(request, 'auth', None), APIGatewayAuthenticationDetails):
            return base_object.objects.none()

        if not callable(getattr(base_object, 'get_queryset_for_principal', None)):
            raise ValueError(
                f'{base_object} does not implement get_queryset_for_principal'
            )
        return base_object.get_queryset_for_principal(request.auth.principal_identifier)

    def has_permission(self, request, view):
        # we cannot determine permissions ownership on list routes, but rely on
        # `get_queryset_for_principal` to be used to filter the queryset appropriately
        if isinstance(getattr(request, 'auth', None), APIGatewayAuthenticationDetails):
            setattr(request, 'should_limit_to_resource_owning_principal', True)
            return True
        return False

    def has_object_permission(self, request, view, obj):
        if not isinstance(getattr(request, 'auth', None), APIGatewayAuthenticationDetails):
            return False

        is_owned_by = getattr(obj, 'is_owned_by', None)
        if not callable(is_owned_by):
            LOG.warn(f'Unable to determine ownership for {obj}')
            return False

        return is_owned_by(request.auth.principal_identifier)


def HasAnyScope(*required_scopes):

    class HasAnyScopesPermission(permissions.BasePermission):
        """
        A permissions class which enforces that the given request has any of the given scopes.

        """

        message = f'Request must have one of the following scope(s) {" ".join(required_scopes)}'

        def has_permission(self, request, view):
            request_scopes = getattr(getattr(request, 'auth', {}), 'scopes', set())
            return len(set(required_scopes) & request_scopes) > 0

        def has_object_permission(self, request, view, obj):
            return self.has_permission(request, view)

    return HasAnyScopesPermission


def SpecifiedPermission(permission: str):

    class HasSpecifiedPermission(permissions.BasePermission):
        """
        A permissions class which ensures that the principal has the correct permissions
        within the permissions specification.

        """

        message = f'Authenticated principal does not have permission {permission}'

        def has_permission(self, request, view):
            principals_with_permission = get_principals_with_permission(permission)
            if request.auth.principal_identifier in principals_with_permission:
                return True

            if request.auth.principal_identifier.scheme != IdentifierSchemes.CRSID:
                LOG.warn('Can only determine group membership for principals identified by CRSID')
                return False

            # special case for people identified by crsid - check whether they are in a
            # lookup group within our list of identities for permission
            groups_with_permission = get_groups_with_permission(permission)
            lookup_group_ids = set([
                identifier.value for identifier in groups_with_permission
                if identifier.scheme == IdentifierSchemes.LOOKUP_GROUP
            ])

            if not lookup_group_ids:
                return False

            return self.is_in_any_lookup_group(
                request.auth.principal_identifier.value, lookup_group_ids
            )

        def has_object_permission(self, request, view, obj):
            return self.has_permission(request, view)

        def is_in_any_lookup_group(self, crsid: str, group_ids: Set[str]) -> bool:
            """
            Determine whether a person identified by a crsid is a member of any of the lookup
            groups provided. Caches the result for 5 minutes to speed up subsequent responses.

            """
            cache_key = f'{crsid}_in_{",".join(group_ids)}'
            cached_response = cache.get(cache_key)
            if cached_response is not None:
                return cached_response

            is_in_group = False
            try:
                group_list = PersonMethods(
                    get_connection()
                ).getGroups(scheme="crsid", identifier=crsid)
                is_in_group = any(
                    (group.groupid for group in group_list if group.groupid in group_ids)
                )
            except IbisException as err:
                LOG.warn(f'Failed to get Lookup groups for {crsid} due to {err}')
                return False

            cache.set(cache_key, is_in_group, timeout=600)
            return is_in_group

    return HasSpecifiedPermission


def get_permissions_for_request(req: request.Request):
    """
    Returns a list of permissions which the request's principal has been granted.
    The permissions will be key any of the keys from the permissions spec, where the
    principal is included within the principals or groups sections.

    """

    return [
        permission_name for permission_name in get_permission_spec().keys() if
        SpecifiedPermission(permission_name)().has_permission(req, None)
    ] if isinstance(req.auth, APIGatewayAuthenticationDetails) else []



from typing import List, Dict, Set
from django.core.cache import cache
from django.conf import settings
from yaml import safe_load
from geddit import geddit
from identitylib.identifiers import Identifier


PERMISSIONS_CACHE_KEY = '__PERMISSION_CACHE__'


def get_permission_spec() -> Dict[str, Dict[str, List]]:
    """
    Returns the permissions spec - expected to be a mapping of permission name to a list of
    identifiers representing people or service identities who have the given permission.

    """

    cached_entry = cache.get(PERMISSIONS_CACHE_KEY)
    if cached_entry:
        return cached_entry

    permissions_spec = safe_load(geddit(settings.PERMISSIONS_SPECIFICATION_URL))
    # Cache the permissions spec for 12 hours. We expect the application to be redeployed
    # when the permissions spec is updated, so it is safe to cache this for a long period
    # of time.
    cache.set(PERMISSIONS_CACHE_KEY, permissions_spec, timeout=43200)

    return permissions_spec


def get_principals_with_permission(permission_name: str) -> Set[Identifier]:
    """
    Returns a set of Identifiers representing people or service identities which have
    the given permission according to the permissions specification.

    """
    return set(
        map(
            lambda identifier_str: Identifier.from_string(identifier_str, find_by_alias=True),
            get_permission_spec().get(permission_name, {}).get('principals', []),
        )
    )


def get_groups_with_permission(permission_name: str) -> Set[Identifier]:
    """
    Returns a set of Identifiers representing groups which indicate membership gives the
    users or service accounts the specified permission.

    """
    return set(
        map(
            lambda identifier_str: Identifier.from_string(identifier_str, find_by_alias=True),
            get_permission_spec().get(permission_name, {}).get('groups', []),
        )
    )



import os
from setuptools import setup, find_packages


def load_requirements(file: str):
    """
    Load requirements file and return non-empty, non-comment lines with leading and trailing
    whitespace stripped.
    """
    with open(os.path.join(os.path.dirname(__file__), file)) as f:
        return [
            line.strip() for line in f
            if line.strip() != '' and not line.strip().startswith('#')
        ]


setup(
    name='django-ucam-apigatewayauth',
    description='A Django module allow auth based on the headers passed from the API Gateway',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://gitlab.developers.cam.ac.uk/uis/devops/django/api-gateway-auth',
    version='0.0.2',
    license='MIT',
    author='DevOps Division, University Information Services, University of Cambridge',
    author_email='devops@uis.cam.ac.uk',
    packages=find_packages(),
    include_package_data=True,
    install_requires=load_requirements('requirements.txt'),
    classifiers=[
        'Development Status :: 3 - Alpha ',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
)



from django.db import models


class TestModel(models.Model):
    """
    A test model to allow us to test functionality which requires a model instance to be passed in.

    """

    @staticmethod
    def get_queryset_for_principal(principal_identifier):
        return TestModel.objects.filter(
            principal_identifier__iexact=principal_identifier.value,
        )

    name = models.TextField('Name', 'name', primary_key=True)
    is_admin = models.BooleanField('Is Admin', 'isAdmin')
    principal_identifier = models.TextField('Principal identifier')



from unittest.mock import MagicMock, patch
from django.test import TestCase
from identitylib.identifiers import Identifier, IdentifierSchemes
from rest_framework.test import APIRequestFactory
from ucamlookup.ibisclient import IbisException

from .mocks.permissions_spec import override_permission_spec
from .mocks.mock_ibis import MockIbisGroup
from .mocks.models import TestModel

from apigatewayauth.api_gateway_auth import APIGatewayAuthenticationDetails

from apigatewayauth.permissions import (
    Disallowed,
    HasAnyScope,
    SpecifiedPermission,
    IsResourceOwningPrincipal,
    get_permissions_for_request
)


class PermissionsTestCase(TestCase):

    def setUp(self):
        super().setUp()

        self.view = MagicMock()
        self.object = MagicMock()

        self.client_auth_details = APIGatewayAuthenticationDetails(
            Identifier('abc', IdentifierSchemes.CRSID), set()
        )

    def request_with_auth(self, auth):
        request = APIRequestFactory().get('/')
        setattr(request, 'auth', auth)
        return request

    def test_disallowed_permission_always_disallows(self):
        permission = Disallowed()

        for auth in [None, self.client_auth_details]:
            self.assertFalse(permission.has_permission(self.request_with_auth(auth), self.view))
            self.assertFalse(
                permission.has_object_permission(
                    self.request_with_auth(auth), self.view, self.object
                )
            )

    def test_is_resource_owning_principal_returns_true_when_authenticated(self):
        permission = IsResourceOwningPrincipal()

        self.assertFalse(permission.has_permission(self.request_with_auth(None), self.view))
        self.assertFalse(
            permission.has_object_permission(
                self.request_with_auth(None), self.view, self.object
            )
        )

        public_client_request = self.request_with_auth(self.client_auth_details)
        self.assertTrue(permission.has_permission(public_client_request, self.view))
        # the request should be flagged as needing limiting
        self.assertTrue(
            getattr(public_client_request, 'should_limit_to_resource_owning_principal')
        )

    def test_is_resource_owning_principal_users_has_object_permission(self):
        permission = IsResourceOwningPrincipal()

        # should be false as the object passed in doesn't have a 'is_owned_by' property
        self.assertFalse(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details),
                self.view,
                {}
            )
        )

        mock_object = MagicMock()
        mock_object.is_owned_by.return_value = False

        self.assertFalse(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details),
                self.view,
                mock_object
            )
        )
        mock_object.is_owned_by.assert_called_with(
            self.client_auth_details.principal_identifier
        )

        mock_object.is_owned_by.return_value = True
        self.assertTrue(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details),
                self.view,
                mock_object
            )
        )

    def test_get_queryset_for_principal_bails_for_unhandled_request(self):
        # bails early if request doesn't require filtering
        self.assertQuerysetEqual(
            IsResourceOwningPrincipal.get_queryset_for_principal(
                self.request_with_auth(self.client_auth_details),
                TestModel
            ),
            TestModel.objects.all()
        )

    def test_get_queryset_for_principal_returns_filtered_queryset(self):
        permission = IsResourceOwningPrincipal()

        request = self.request_with_auth(self.client_auth_details)

        # pass through `has_permission` to mark `should_limit_to_resource_owning_principal`
        # on the requests
        self.assertTrue(permission.has_permission(request, self.view))

        self.assertQuerysetEqual(
            IsResourceOwningPrincipal.get_queryset_for_principal(request, TestModel),
            TestModel.get_queryset_for_principal(
                self.client_auth_details.principal_identifier
            )
        )

    def test_get_queryset_for_principal_returns_no_objects_if_no_request_auth(self):
        permission = IsResourceOwningPrincipal()

        request = APIRequestFactory().get('/')

        # `has_permission` should return false as we have no auth details on the request
        self.assertFalse(permission.has_permission(request, self.view))

        # Set should_limit_to_resource_owning_principal manually, which allows us to hit the
        # condition of having no principal whilst trying to get the queryset for a principal.
        # This should never really happen, but we handle the condition in the permission class
        # so it's worth a test.
        setattr(request, 'should_limit_to_resource_owning_principal', True)

        # If we get the queryset for the principal we should get objects.none() as we do not have
        # a principal.
        self.assertQuerysetEqual(
            IsResourceOwningPrincipal.get_queryset_for_principal(request, TestModel),
            TestModel.objects.none()
        )

    def test_get_queryset_for_principal_throws_on_an_unknown_model(self):
        permission = IsResourceOwningPrincipal()

        request = self.request_with_auth(self.client_auth_details)
        self.assertTrue(permission.has_permission(request, self.view))

        # using get_queryset_for_principal without a model defining get_queryset_for_principal
        # should cause a value error with a desciption message
        with self.assertRaisesRegex(
            ValueError, '{} does not implement get_queryset_for_principal'
        ):
            IsResourceOwningPrincipal.get_queryset_for_principal(request, {})

    def test_has_any_scope(self):
        permission = HasAnyScope('a', 'b.readonly')()

        for auth in [None, self.client_auth_details]:
            self.assertFalse(permission.has_permission(self.request_with_auth(auth), self.view))
            self.assertFalse(
                permission.has_object_permission(
                    self.request_with_auth(auth), self.view, self.object
                )
            )

        self.assertTrue(
            permission.has_permission(
                self.request_with_auth(
                    APIGatewayAuthenticationDetails(
                        Identifier('abc', IdentifierSchemes.CRSID),
                        set(['b.readonly'])
                    )
                ),
                self.view,
            )
        )

        self.assertFalse(
            permission.has_object_permission(
                self.request_with_auth(
                    APIGatewayAuthenticationDetails(
                        Identifier('abc', IdentifierSchemes.CRSID),
                        set(['abcd'])
                    )
                ),
                self.view,
                self.object,
            )
        )

    @patch('apigatewayauth.permissions.get_principals_with_permission')
    def test_has_specified_permission_return_false_if_not_matching(self, get_identities_mock):
        get_identities_mock.return_value = set([
            Identifier('1000', IdentifierSchemes.STAFF_NUMBER),
            Identifier('abc123', IdentifierSchemes.CRSID)
        ])
        permission = SpecifiedPermission('READ_BOOKS')()

        self.assertFalse(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        self.assertFalse(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details), self.view, {}
            )
        )
        get_identities_mock.assert_called_with('READ_BOOKS')

    @patch('apigatewayauth.permissions.get_principals_with_permission')
    def test_has_specified_permission_return_true_if_matching(self, get_identities_mock):
        get_identities_mock.return_value = set([
            Identifier('1000', IdentifierSchemes.STAFF_NUMBER),
            self.client_auth_details.principal_identifier
        ])
        permission = SpecifiedPermission('READ_BOOKS')()

        self.assertTrue(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        self.assertTrue(
            permission.has_object_permission(
                self.request_with_auth(self.client_auth_details), self.view, {}
            )
        )
        get_identities_mock.assert_called_with('READ_BOOKS')

    @patch('apigatewayauth.permissions.get_principals_with_permission')
    @patch('apigatewayauth.permissions.get_groups_with_permission')
    @patch('apigatewayauth.permissions.PersonMethods.getGroups')
    def test_does_not_query_lookup_groups_for_non_crsid_principal(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set([
            Identifier('abc123', IdentifierSchemes.CRSID),
            Identifier('abc333', IdentifierSchemes.CRSID)
        ])
        get_groups_with_permission_mock.return_value = set([
            Identifier('1000', IdentifierSchemes.LOOKUP_GROUP),
        ])
        permission = SpecifiedPermission('READ_MAGAZINES')()

        auth_details = APIGatewayAuthenticationDetails(
            Identifier('1', IdentifierSchemes.STAFF_NUMBER),
            set([])
        )

        self.assertFalse(
            permission.has_permission(self.request_with_auth(auth_details), self.view)
        )
        self.assertFalse(
            permission.has_object_permission(self.request_with_auth(auth_details), self.view, {})
        )
        get_identities_mock.assert_called_with('READ_MAGAZINES')

        # should not have queried for groups as our principal does not have a crsid identifier
        get_groups_mock.assert_not_called()

    @patch('apigatewayauth.permissions.get_principals_with_permission')
    @patch('apigatewayauth.permissions.get_groups_with_permission')
    @patch('apigatewayauth.permissions.PersonMethods.getGroups')
    def test_does_not_query_lookup_groups_if_no_lookup_groups_in_permissions_spec(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set([
            Identifier('1000', IdentifierSchemes.STAFF_NUMBER),
            Identifier('bb123', IdentifierSchemes.CRSID)
        ])
        get_groups_with_permission_mock.return_value = set()
        permission = SpecifiedPermission('READ_PAMPHLETS')()

        self.assertFalse(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_identities_mock.assert_called_with('READ_PAMPHLETS')

        # should not have queried for groups as we do not have a lookup group in our
        # allowed identities list
        get_groups_mock.assert_not_called()

    @patch('apigatewayauth.permissions.get_principals_with_permission')
    @patch('apigatewayauth.permissions.get_groups_with_permission')
    @patch('apigatewayauth.permissions.PersonMethods.getGroups')
    def test_will_return_true_if_principal_is_in_specified_lookup_group(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set([
            Identifier('cc123', IdentifierSchemes.CRSID)
        ])
        get_groups_with_permission_mock.return_value = set([
            Identifier('1001', IdentifierSchemes.LOOKUP_GROUP)
        ])
        permission = SpecifiedPermission('READ_THE_NEWS')()
        get_groups_mock.return_value = [MockIbisGroup('1001'), MockIbisGroup('1002')]

        self.assertTrue(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_identities_mock.assert_called_with('READ_THE_NEWS')

        # should have queried Lookup for the principal's groups
        get_groups_mock.assert_called_with(
            scheme='crsid', identifier=self.client_auth_details.principal_identifier.value
        )

        # querying again returns true without going to Lookup due to caching the result
        get_groups_mock.reset_mock()
        self.assertTrue(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_groups_mock.assert_not_called()

    @patch('apigatewayauth.permissions.get_principals_with_permission')
    @patch('apigatewayauth.permissions.get_groups_with_permission')
    @patch('apigatewayauth.permissions.PersonMethods.getGroups')
    def test_will_return_false_if_principal_is_not_in_specified_lookup_group(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set([
            Identifier('dd123', IdentifierSchemes.CRSID)
        ])
        get_groups_with_permission_mock.return_value = set([
            Identifier('10101', IdentifierSchemes.LOOKUP_GROUP)
        ])
        permission = SpecifiedPermission('READ_THE_ROOM')()
        get_groups_mock.return_value = [MockIbisGroup('1001'), MockIbisGroup('1002')]

        self.assertFalse(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_identities_mock.assert_called_with('READ_THE_ROOM')

        # should have queried Lookup for the principal's groups
        get_groups_mock.assert_called_with(
            scheme='crsid', identifier=self.client_auth_details.principal_identifier.value
        )

    @patch('apigatewayauth.permissions.get_principals_with_permission')
    @patch('apigatewayauth.permissions.get_groups_with_permission')
    @patch('apigatewayauth.permissions.PersonMethods.getGroups')
    def test_will_return_false_if_unable_to_read_lookup_groups(
        self,
        get_groups_mock,
        get_groups_with_permission_mock,
        get_identities_mock,
    ):
        get_identities_mock.return_value = set([
            Identifier('ee123', IdentifierSchemes.CRSID),
        ])
        get_groups_with_permission_mock.return_value = set([
            Identifier('2000', IdentifierSchemes.LOOKUP_GROUP),
        ])
        permission = SpecifiedPermission('READ_THE_FUTURE')()

        def throw_ibis_error(*args, **kwargs):
            error = ValueError('failed')
            setattr(error, 'message', 'failed')  # ibis exception needs a 'message' attribute
            raise IbisException(error)

        get_groups_mock.side_effect = throw_ibis_error

        self.assertFalse(
            permission.has_permission(self.request_with_auth(self.client_auth_details), self.view)
        )
        get_identities_mock.assert_called_with('READ_THE_FUTURE')

        # should have queried Lookup for the principal's groups
        get_groups_mock.assert_called_with(
            scheme='crsid', identifier=self.client_auth_details.principal_identifier.value
        )

    @override_permission_spec({
        'DATA_READER': {
            'principals': {
                str(Identifier('abc44', IdentifierSchemes.CRSID)),
                str(Identifier('abc55', IdentifierSchemes.CRSID))
            }
        },
        'DATA_OWNER': {
            'principals': {
                str(Identifier('abc44', IdentifierSchemes.CRSID))
            }
        }
    })
    def test_get_permissions_for_request(self):
        """
        Test that `get_permissions_for_request` returns correctly based on the user authenticated
        on the request.

        """

        mock_request = APIRequestFactory().get('/')
        setattr(mock_request, 'auth', None)

        # expect an empty list to be returned if no auth details provided
        self.assertListEqual(get_permissions_for_request(mock_request), [])

        # if we have a principal not in our permissions spec we should get an empty list
        mock_request.auth = APIGatewayAuthenticationDetails(
            Identifier('abc123', IdentifierSchemes.CRSID), []
        )
        self.assertListEqual(get_permissions_for_request(mock_request), [])

        # abc55 should be shown as a `DATA_READER`
        mock_request.auth = APIGatewayAuthenticationDetails(
            Identifier('abc55', IdentifierSchemes.CRSID), []
        )
        self.assertListEqual(get_permissions_for_request(mock_request), ['DATA_READER'])

        # abc44 should be shown as a `DATA_READER` and `DATA_WRITER`
        mock_request.auth = APIGatewayAuthenticationDetails(
            Identifier('abc44', IdentifierSchemes.CRSID), []
        )
        self.assertListEqual(
            get_permissions_for_request(mock_request), ['DATA_OWNER', 'DATA_READER']
        )
